@ mode6.asm
@ mic, 2005 | micol972@gmail.com

@ Draws a 256-color plasma on BG2 in mode 6

.text
.global main
.global forever


.equ PALETTE_RAM,	0x5000000
.equ PALETTE_RAM_SIZE,	0x0000400
.equ BG_RAM,		0x6000000
.equ BG_SUB_RAM, 	0x6200000

.equ DISPCNT,		0x4000000
.equ BG2CNT,		0x400000c
.equ BG2PA,		0x4000020
.equ BG2X,		0x4000028
.equ POWERCNT,		0x4000304
.equ DISPCNT_SUB,	0x4001000
.equ BG0CNT_SUB,	0x4001008
.equ BG0HOFS_SUB,	0x4001010
.equ BG0VOFS_SUB,	0x4001012

.equ BG_MODE_0,		0x10000
.equ BG_MODE_6,		0x10006

.equ BG0_ENABLE,	0x0100
.equ BG1_ENABLE,	0x0200
.equ BG2_ENABLE,	0x0400
.equ BG3_ENABLE,	0x0800
.equ OBJ_ENABLE,	0x1000

.equ BANK_ENABLE,	0x80
.equ BANK_BG_6000000,	0x01


main:

@ Write some junk to DISPCNT (just for fun..)
ldr r0,=0x11340
mov r1,#DISPCNT
str r0,[r1]


@ Turn on both screens, main screen on top
ldr r0,=0x830F
ldr r1,=POWERCNT
str r0,[r1]


@ Map DTCM to 0x800000, 16kB
mov r0,#0x800000
orr r0,r0,#0xA
mcr 15,0,r0,c9,c1,0


@ Map all VRAM banks as BGs
mov r0,#(BANK_ENABLE|BANK_BG_6000000)
ldr r1,=0x4000240
strb r0,[r1],#1
add r0,r0,#8
strb r0,[r1],#1
add r0,r0,#8
strb r0,[r1],#1
add r0,r0,#8
strb r0,[r1],#1


@ Set up palette
mov r0,#0
mov r1,#PALETTE_RAM
add r9,r1,#PALETTE_RAM_SIZE
mov r7,#256
mov r8,#384
set_palette:
	mov r2,r0,lsr#1		@ 0..31
	rsb r3,r2,#31		@ 31..0

	strh r2,[r1]
	strh r2,[r9]
	orr r4,r3,r2,lsl#5
	strh r4,[r1,#128]
	strh r4,[r9,#128]
	mov r4,r2,lsl#10
	orr r4,r4,r3,lsl#5
	strh r4,[r1,r7]
	strh r4,[r9,r7]
	mov r4,r3,lsl#10
	strh r4,[r1,r8]
	strh r4,[r9,r8]
	add r1,r1,#2
	add r9,r9,#2
	add r0,r0,#1
	cmp r0,#64
	bne set_palette

mov r1,#PALETTE_RAM
add r1,r1,#122
ldr r0,=special_colors
mov r2,#3
set_special_colors:
	ldrh r3,[r0]
	strh r3,[r1]
	ldrh r3,[r0,#6]
	strh r3,[r1,#128]
	ldrh r3,[r0,#12]
	strh r3,[r1,r7]
	add r0,r0,#2
	add r1,r1,#2
	subs r2,r2,#1
	bne set_special_colors
	


@ Set up rotation/scaling parameters for BG2
ldr r0,=BG2PA
mov r2,#0x20		@ 0.125
mov r1,#0x100		@ 1.00
strh r1,[r0],#2
strh r2,[r0],#2
strh r2,[r0],#2
mov r1,#0xC0		@ 0.75
strh r1,[r0],#2
mov r2,#0
str r2,[r0],#4
str r2,[r0],#4


@ Set BG mode 6 and enable BG2
ldr r0,=(BG_MODE_6 | BG2_ENABLE)
mov r1,#DISPCNT
str r0,[r1]	

@ Enable screen wrapping
mov r0,#0x2000
ldr r1,=BG2CNT
strh r0,[r1]


@ Store some variables for the plasma at offset 0x100000 in RAM
ldr r0,=0x2100000
mov r8,#0
mov r9,#0
mov r10,#0
mov r11,#0
mov r12,#0
stmia r0!,{r8-r12}
stmia r0!,{r8-r12}



@@@@ Sub core setup @@@@

@ Copy font data
ldr r0,=font
ldr r1,=BG_SUB_RAM
ldr r2,=2688
copy_font:
	ldrh r3,[r0],#2
	strh r3,[r1],#2
	subs r2,r2,#1
	bne copy_font
	
@ Set mode 0 and enable BG0
ldr r0,=(BG_MODE_0 | BG0_ENABLE)
ldr r1,=DISPCNT_SUB
str r0,[r1]	

@ Sub BG0: text bg, 256 colors, 256x256 pixels, screen base 0x4000, priority 0
ldr r0,=0x0880
ldr r1,=BG0CNT_SUB
strh r0,[r1]

@ Scroll Sub BG0 to the right a bit
ldr r0,=0xFFFC
ldr r1,=BG0HOFS_SUB
strh r0,[r1]

@ Print some text on Sub BG0
bl VSync
ldr r0,=szText
mov r1,#80
mov r2,#88
ldr r3,=(BG_SUB_RAM + 0x4000 - BG_RAM)
bl DrawText



forever:
	bl VSync

	@ Load trig table offsets 3 and 4
	ldr r12,=0x210000C
	ldr r11,[r12],#-4
	ldr r10,[r12],#-4

	mov r0,#BG_RAM
	ldr r1,=trigtable
	mov r2,#192
	yloop:
		add r8,r0,#512
		sub r8,r8,#2
		add r10,r10,#1
		add r11,r11,#2
		and r10,r10,#0xFF
		and r11,r11,#0xFF
		
		@ Load trig table offset 1 and 2
		ldr r6,[r12,#-4]
		ldr r7,[r12]
		mov r3,#128
		xloop:
			add r6,r6,#2
			add r7,r7,#3
			and r6,r6,#0xFF
			and r7,r7,#0xFF
			ldrb r4,[r1,r6]
			ldrb r5,[r1,r7]
			add r4,r4,r5
			ldrb r5,[r1,r10]
			add r4,r4,r5
			ldrb r5,[r1,r11]
			add r4,r4,r5
			and r4,r4,#0xFF
			orr r4,r4,r4,lsl#8
			strh r4,[r0],#2
			mvn r4,r4
			strh r4,[r8],#-2 
			subs r3,r3,#1
			bne xloop
		add r0,r0,#256
		subs r2,r2,#1
		bne yloop

	@ Update plasma variables
	ldr r0,=0x2100000
	ldr r1,[r0]
	sub r1,r1,#3
	str r1,[r0],#4

	ldr r1,[r0]
	add r1,r1,#2
	str r1,[r0],#4

	ldr r1,[r0]
	sub r1,r1,#3
	str r1,[r0],#4

	ldr r1,[r0]
	add r1,r1,#1
	str r1,[r0],#4

	ldr r1,[r0]
	add r1,r1,#0xC0		@ Scroll 0.75 pixels
	str r1,[r0]
	ldr r0,=BG2X
	str r1,[r0]	
	
	b forever




VSync:
	stmfd sp!,{r0-r1}
	mov r0,#DISPCNT
	add r0,r0,#4		@ DISPCNT+4 = DISPSTAT

	vs_loop1:
    		ldrh r1,[r0]
    		ands r1,r1,#1
  	bne vs_loop1

  	vs_loop2:
    		ldrh r1,[r0]
    		ands r1,r1,#1
  	beq vs_loop2

  	ldmfd sp!,{r0-r1}
  	mov pc,lr


DrawText:
@ r0: lpszText
@ r1: x
@ r2: y
@ r3: vram offset
	stmfd sp!,{r4-r10}

	mov r9,r2,lsr#3
	mov r9,r9,lsl#6
	mov r7,r1,lsr#3
	add r9,r9,r3
	add r9,r9,r7,lsl#1
	add r9,r9,#0x6000000
	
dt_cloop:	
	ldrb r4,[r0],#1
	cmp r4,#0
	beq dt_null
	cmp r4,#32
	moveq r4,#0
	subne r4,r4,#37

	strh r4,[r9],#2	
	
	b dt_cloop
dt_null:
	ldmfd sp!,{r4-r10}
	mov pc,lr

.pool
.align

	

.align
.pool

.align 2
font:
	.incbin "font_grad.pat"
.align 2
szText: .asciz "MODE 6 DEMO"

.align 2
special_colors:
.hword 0x463e,0x5eff,0x6b7d
.hword 0x0fd9,0x4bfe,0x6bff
.hword 0x7e10,0x7eb5,0x777a


.align 2
trigtable:
.byte 64,63,63,63,63,63,63,63,63,63,63,62,62,62,62,61
.byte 61,61,60,60,60,59,59,58,58,58,57,57,56,56,55,55
.byte 54,53,53,52,52,51,50,50,49,49,48,47,46,46,45,44
.byte 44,43,42,41,41,40,39,38,38,37,36,35,34,34,33,32
.byte 31,31,30,29,28,27,27,26,25,24,24,23,22,21,20,20
.byte 19,18,18,17,16,15,15,14,13,13,12,12,11,10,10,9
.byte 9,8,8,7,7,6,6,5,5,4,4,3,3,3,2,2
.byte 2,2,1,1,1,1,0,0,0,0,0,0,0,0,0,0
.byte 0,0,0,0,0,0,0,0,0,0,1,1,1,1,2,2
.byte 2,2,3,3,3,4,4,5,5,6,6,7,7,8,8,9
.byte 9,10,10,11,12,12,13,13,14,15,16,16,17,18,18,19
.byte 20,20,21,22,23,24,24,25,26,27,27,28,29,30,31,31
.byte 32,33,34,34,35,36,37,38,38,39,40,41,41,42,43,44
.byte 44,45,46,46,47,48,49,49,50,50,51,52,52,53,53,54
.byte 55,55,56,56,57,57,58,58,58,59,59,60,60,60,61,61
.byte 61,62,62,62,62,63,63,63,63,63,63,63,63,63,63,64
  

.align
.pool
.end
